/**
 * Copyright (C) 2007-2012 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef LTOP_LOGIN_BOX_H
#define LTOP_LOGIN_BOX_H

#include <gtk/gtk.h>

typedef enum
{
    LTOP_LOGIN_INFO_DATA_TYPE_IMAGE = 0,
    LTOP_LOGIN_INFO_DATA_TYPE_CONFIG = 1,
} LtopLoginInfoDataType;

typedef struct _LtopLoginInfoData LtopLoginInfoData;
struct _LtopLoginInfoData
{
    LtopLoginInfoDataType type;
    gboolean available;
    gchar *id;
    gchar *version;
    gchar *text;
    gchar *desc;
};

LtopLoginInfoData *ltop_login_info_data_new ();

void ltop_login_info_data_free (LtopLoginInfoData *data);


typedef struct _LtopLoginInfo LtopLoginInfo;
struct _LtopLoginInfo
{
    gboolean valid;

    GList *images;
    GList *configs;
};

LtopLoginInfo *ltop_login_info_new ();

void ltop_login_info_free (LtopLoginInfo *li);



typedef struct _LtopLoginBox        LtopLoginBox;
typedef struct _LtopLoginBoxClass   LtopLoginBoxClass;
typedef struct _LtopLoginBoxPrivate LtopLoginBoxPrivate;

#define LTOP_TYPE_LOGIN_BOX          (ltop_login_box_get_type())
#define LTOP_LOGIN_BOX(obj)          G_TYPE_CHECK_INSTANCE_CAST (obj, ltop_login_box_get_type (), LtopLoginBox)
#define LTOP_LOGIN_BOX_CLASS(klass)  G_TYPE_CHECK_CLASS_CAST (klass, ltop_login_box_get_type (), LtopLoginBoxClass)
#define LTOP_IS_LOGIN_BOX(obj)       G_TYPE_CHECK_INSTANCE_TYPE (obj, ltop_login_box_get_type ())

struct _LtopLoginBoxClass
{
    GtkTableClass __parent__;
};

struct _LtopLoginBox
{
    GtkTable __parent__;

    LtopLoginBoxPrivate *priv;
};

typedef void (*LtopLoginFunc) (LtopLoginBox *lb,
                               const gchar *str,
                               gpointer data);

GType ltop_login_box_get_type (void) G_GNUC_CONST;

GtkWidget *ltop_login_box_new (const gchar *email,
                               const gchar *invoice);

void ltop_login_box_set_email   (LtopLoginBox *lb,
                                 const gchar *email);
void ltop_login_box_set_invoice (LtopLoginBox *lb,
                                 const gchar *invoice);

const gchar *ltop_login_box_get_email   (LtopLoginBox *lb);
const gchar *ltop_login_box_get_invoice (LtopLoginBox *lb);

LtopLoginInfo *ltop_login_box_validate_login (const gchar *email,
                                              const gchar *invoice);

LtopLoginInfo *ltop_login_box_login       (LtopLoginBox *lb);
void           ltop_login_box_login_async (LtopLoginBox *lb,
                                           LtopLoginFunc func,
                                           gpointer data);

gboolean ltop_login_box_check_connected ();

gboolean ltop_login_box_apply_config (const gchar *config_id, const gchar *config,
                                      const gchar *email, const gchar *invoice,
                                      gchar **output);

void ltop_login_box_register_download (const gchar *id,
                                       guint type,
                                       const gchar *email,
                                       const gchar *invoice);

#endif /* ! LTOP_LOGIN_BOX_H */

