/**
 * Copyright (C) 2007-2012 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif

#include <glib.h>
#include <glib/gstdio.h>
#include <gtk/gtk.h>

#include <libxfce4util/libxfce4util.h>

#include <libxfcegui4.h>

const gchar *conffile;

GList *pids = NULL;

gboolean use_pw = FALSE;
gboolean view_only = FALSE;
gboolean allow_multi = FALSE;
gboolean autostart = FALSE;

GtkWidget *window;
GtkWidget *use_pw_chk;
GtkWidget *view_only_chk;
GtkWidget *allow_multi_chk;
GtkWidget *entry1, *entry2;
GtkWidget *autostart_chk;
GtkWidget *stop_btn, *start_btn, *restart_btn;

enum {
    OK = 0,
    CANCEL
};

enum {
    SERVER_START = 0,
    SERVER_STOP,
    SERVER_RESTART,
};

const gchar *default_options =
    "xkb\n"
    "forever\n"
    "loop\n";

/* helpers */
static const gchar *
get_conffile ()
{
    gchar *tmp = g_strconcat (getenv ("HOME"), "/.x11vncrc", NULL);
    const gchar *ret = g_strdup (tmp);
    g_free (tmp);
    return ret;
}

static void
get_server_pids ()
{
    GDir *dir;
    gchar filename[64];
    gchar *buffer = NULL;
    gchar procname[64], junk[1024];
    gchar **spl = NULL;
    int curpid, curuid, i = 0;
    int idummy;
    const gchar *tmp;

    if (pids)
    {
        g_list_free (pids);
        pids = NULL;
    }

    dir = g_dir_open ("/proc", 0, NULL);
    if (!dir)
        g_error ("No /proc dir, something is going really wrong.");

    while ((tmp = g_dir_read_name (dir)) != NULL)
    {
        if (strspn (tmp, "0123456789") != strlen (tmp))
            continue;

        /* parse the stat file to get the process name */
        g_sprintf (filename, "/proc/%s/stat", tmp);
        if (g_access (filename, R_OK))
            continue;

        if (!g_file_get_contents (filename, &buffer, NULL, NULL))
            continue;

        sscanf (buffer, "%i (%255s %s", &curpid, &procname[0], &junk[0]);
        g_free (buffer);

        if (strncmp ("x11vnc)", procname, strlen (procname)))
            continue;

        /* does the process belong to the user? */
        g_sprintf (filename, "/proc/%s/status", tmp);
        if (g_access (filename, R_OK))
            continue;

        if (!g_file_get_contents (filename, &buffer, NULL, NULL))
            continue;

        spl = g_strsplit (buffer, "\n", 0);
        curuid = -1;
        while (spl[i])
        {
            if (!strncmp (spl[i], "Uid:", 4))
            {
                sscanf (spl[i], "%s %d %d %d %d", junk, &curuid,
                        &idummy, &idummy, &idummy);
                if (curuid != getuid ())
                {
                    curuid = -1;
                    continue;
                }
                break;
            }
            i++;
        }
        g_strfreev (spl);

        if (curuid == -1)
            continue;

        pids = g_list_append (pids, GINT_TO_POINTER (curpid));
    }

    g_dir_close (dir);
}

static void
set_startup ()
{
    gchar file[64];
    gchar dir[64];
    const gchar *home = xfce_get_homedir ();

    g_snprintf (file, 64, "%s/.config/autostart/vncserver.desktop", home);
    g_snprintf (dir, 64, "%s/.config/autostart/", home);
    mkdir (dir, 0755);

    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (autostart_chk)))
    {
        const gchar *contents =
            "[Desktop Entry]\n"
            "Name=VNC Server\n"
            "Exec=x11vnc\n"
            "Type=Application\n"
            "X-GNOME-Autostart-Delay=30\n";
        g_file_set_contents (file, contents, -1, NULL);
    } else {
        if (!g_access (file, W_OK))
            unlink (file);
    }
}

static void
write_config ()
{
    use_pw = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (use_pw_chk));
    view_only = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (view_only_chk));
    allow_multi = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (allow_multi_chk));

    gchar *contents = g_strconcat (
        default_options,
        use_pw      ? "usepw\n" : "",
        view_only   ? "viewonly\n" : "",
        allow_multi ? "shared\n" : "",
        NULL);

    g_file_set_contents (conffile, contents, -1, NULL);

    set_startup ();
}

/* from libxfcegui4
 * Copyright (c) 2003,2005 Jasper Huijsmans <jasper@xfce.org>
 * Copyright (c) 2003 Benedikt Meurer <benny@xfce.org>
 */
GtkWidget*
xfce_create_mixed_button (const gchar *stock, const gchar *text)
{
    GtkWidget *button, *align, *image, *hbox, *label;

    button = gtk_button_new ();
    label = gtk_label_new_with_mnemonic (text);
    gtk_label_set_mnemonic_widget (GTK_LABEL (label), button);

    image = gtk_image_new_from_stock (stock, GTK_ICON_SIZE_BUTTON);
    hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 2);

    align = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);

    gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
    gtk_box_pack_end (GTK_BOX (hbox), label, FALSE, FALSE, 0);

    gtk_container_add (GTK_CONTAINER (button), align);
    gtk_container_add (GTK_CONTAINER (align), hbox);
    gtk_widget_show_all (align);

    return button;
}

GtkWidget *
xfce_create_mixed_button_from_name (const gchar *name, const gchar *text)
{
    GtkWidget *button, *align, *image, *hbox, *label;

    button = gtk_button_new ();
    label = gtk_label_new_with_mnemonic (text);
    gtk_label_set_mnemonic_widget (GTK_LABEL (label), button);

    image = gtk_image_new_from_icon_name (name, GTK_ICON_SIZE_BUTTON);
    hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 2);

    align = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);

    gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
    gtk_box_pack_end (GTK_BOX (hbox), label, FALSE, FALSE, 0);

    gtk_container_add (GTK_CONTAINER (button), align);
    gtk_container_add (GTK_CONTAINER (align), hbox);
    gtk_widget_show_all (align);

    return button;
}

/* callbacks */
static void
on_pw_chk_toggled (GtkWidget *toggle, gpointer btn)
{
    gboolean active = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (toggle));
    use_pw = active;
    gtk_widget_set_sensitive (GTK_WIDGET (btn), active);
}

static void
pwd_ok (GtkWidget *entry, gpointer data)
{
    gchar *t1, *t2;
    GtkWidget *ok = GTK_WIDGET (data);

    t1 = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry1)));
    t2 = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry2)));

    if (strlen (t1) && (strcmp (t1, t2) == 0))
        gtk_widget_set_sensitive (ok, TRUE);
    else
        gtk_widget_set_sensitive (ok, FALSE);

    g_free(t1);
    g_free(t2);
}

static void
set_passwd (gchar *pwd)
{
    /*
     * setting the password seems to hang if a config exists (with 'loop' in
     * it; so store the config, set the passwd and restore the config
     */
    gchar *conffile = g_strdup (get_conffile ());
    gchar *tmpfile = g_strconcat (conffile, ".bak", NULL);
    rename (conffile, tmpfile);

    gchar *tmp = g_strconcat (xfce_get_homedir (), "/.vnc", NULL);
    mkdir (tmp, 0755);
    gchar *cmd = g_strconcat ("x11vnc -storepasswd ", pwd, " ", tmp, "/passwd", NULL);
    g_spawn_command_line_sync (cmd, NULL, NULL, NULL, NULL);

    rename (tmpfile, conffile);

    g_free (tmp);
    g_free (cmd);
    g_free (tmpfile);
}

static void
on_pw_btn_clicked (GtkWidget *btn, gpointer data)
{
    GtkWidget *dialog, *ok, *cancel;
    GtkWidget *label, *table;

    int response;

    gchar *pwd = NULL;

    dialog = gtk_dialog_new ();
    gtk_window_set_title (GTK_WINDOW (dialog), _("Set the password"));
    gtk_window_set_destroy_with_parent (GTK_WINDOW (window), TRUE);
    gtk_window_set_modal (GTK_WINDOW (window), TRUE);

    table = gtk_table_new (2, 2, FALSE);
    gtk_container_add (GTK_CONTAINER (
        gtk_dialog_get_content_area (GTK_DIALOG (dialog))),
        table);

    entry1 = gtk_entry_new ();
    gtk_entry_set_visibility (GTK_ENTRY (entry1), FALSE);
    gtk_table_attach (GTK_TABLE (table), entry1, 1, 2, 1, 2,
             GTK_SHRINK, GTK_SHRINK, 2, 2);
    entry2 = gtk_entry_new ();
    gtk_entry_set_visibility (GTK_ENTRY (entry2), FALSE);
    gtk_table_attach (GTK_TABLE (table), entry2, 1, 2, 2, 3,
             GTK_SHRINK, GTK_SHRINK, 2, 2);
    label = gtk_label_new (_("Password: "));
    gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
             GTK_SHRINK, GTK_SHRINK, 2, 2);
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    label = gtk_label_new (_("Retype password: "));
    gtk_table_attach (GTK_TABLE (table), label, 0, 1, 2, 3,
             GTK_SHRINK, GTK_SHRINK, 2, 2);
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

    cancel = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
    ok = gtk_button_new_from_stock (GTK_STOCK_OK);
    gtk_widget_set_sensitive (ok, FALSE);
    gtk_dialog_add_action_widget (GTK_DIALOG (dialog), cancel, CANCEL);
    gtk_dialog_add_action_widget (GTK_DIALOG (dialog), ok, OK);

    g_signal_connect (G_OBJECT (entry1), "changed",
             G_CALLBACK (pwd_ok), ok);
    g_signal_connect (G_OBJECT (entry2), "changed",
             G_CALLBACK (pwd_ok), ok);

    gtk_widget_show_all (dialog);
    response = gtk_dialog_run (GTK_DIALOG (dialog));
    if (response == OK) {
        pwd = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry1)));
        if (pwd)
        {
            set_passwd (pwd);
            g_free (pwd);
        }
    }

    gtk_entry_set_text (GTK_ENTRY (entry1), "");
    gtk_entry_set_text (GTK_ENTRY (entry2), "");
    gtk_widget_destroy(dialog);
}

static void
set_action_buttons_visibility ()
{
    gboolean have_pid = FALSE;

    get_server_pids ();
    if (g_list_length (pids) > 0)
        have_pid = TRUE;

    gtk_widget_set_sensitive (start_btn, ! have_pid);
    gtk_widget_set_sensitive (restart_btn, have_pid);
    gtk_widget_set_sensitive (stop_btn, have_pid);
}

static gboolean
set_action_buttons_visibility_timeout (gpointer data)
{
    set_action_buttons_visibility ();
    return FALSE;
}

static void
kill_pid (gpointer data, gpointer unused)
{
    int pid = GPOINTER_TO_INT (data);
    kill (pid, SIGTERM);
}

static void
on_server_action_clicked (GtkButton *button, gpointer data)
{
    int action = GPOINTER_TO_INT (g_object_get_data (
                    G_OBJECT (button), "id"));

    switch (action) {
        case SERVER_STOP:
            g_list_foreach (pids, kill_pid, NULL);
            break;
        case SERVER_RESTART:
            g_list_foreach (pids, kill_pid, NULL);
        case SERVER_START:
            write_config ();
            g_spawn_command_line_async ("x11vnc", NULL);
            break;
        default:
            g_warning ("unknown action button id: %d", action);
    }

    /* let some time to x11vnc to fork/die */
    g_timeout_add_seconds (1, set_action_buttons_visibility_timeout, NULL);
}

static void
quit (GtkDialog *dialog, gint response, gpointer data)
{
    write_config ();
    gtk_main_quit ();
}

int
main (int argc, gchar **argv)
{
    /* FIXME: forbid to run this as root */
    gchar *contents = NULL;

    GtkWidget *heading;
    GtkWidget *content_area;
    GtkWidget *mhbox;
    GtkWidget *vbox, *hbox, *bbox;
    GtkWidget *frame;
    GtkWidget *vs;
    GtkWidget *set_passwd_btn;

    xfce_textdomain(GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR, "UTF-8");
    gtk_init (&argc, &argv);

    conffile = get_conffile ();
    if (g_file_get_contents (conffile, &contents, NULL, NULL))
    {
        gchar **split = g_strsplit (contents, "\n", 0);
        int length = g_strv_length (split);
        int i;
        for (i=0;i<length;i++)
        {
            if (!strcmp ("usepw", split[i]))
            {
                use_pw = TRUE;
                continue;
            }
            if (!strcmp ("viewonly", split[i]))
            {
                view_only = TRUE;
                continue;
            }
            if (!strcmp ("shared", split[i]))
            {
                allow_multi = TRUE;
                continue;
            }
        }
        g_strfreev (split);
        g_free (contents);
    }

    /* autostart config */
    gchar tmp[128];
    g_sprintf (tmp, "%s/.config/autostart/vncserver.desktop", xfce_get_homedir ());
    if (g_access (tmp, R_OK) == 0)
        autostart = TRUE;

    /* build the GUI */
    gtk_window_set_default_icon_name ("gnome-remote-desktop");
    window = gtk_dialog_new_with_buttons (_("VNC remote access configuration"),
                                            NULL,
                                            0,
                                            GTK_STOCK_QUIT,
                                            GTK_RESPONSE_ACCEPT,
                                            NULL);
    content_area = gtk_dialog_get_content_area (GTK_DIALOG (window));
    gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
    g_signal_connect (G_OBJECT (window), "response",
        G_CALLBACK (quit), NULL);

    heading = ltop_heading_new ();
    ltop_heading_set_title (LTOP_HEADING (heading),
        _("Configue the VNC remote access service"));
    ltop_heading_set_icon_name (LTOP_HEADING (heading), "gnome-remote-desktop");
    gtk_box_pack_start (GTK_BOX (content_area), heading, FALSE, FALSE, 0);

    mhbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
    gtk_box_pack_start (GTK_BOX (content_area), mhbox, FALSE, FALSE, 0);

    /* config */

    vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 2);
    vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 2);

    autostart_chk = gtk_check_button_new_with_label (_(
        "Automatically start the server"));
    gtk_box_pack_start (GTK_BOX (vbox), autostart_chk, FALSE, FALSE, 0);

    hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
    use_pw_chk = gtk_check_button_new_with_label (_("Request a password"));
    gtk_box_pack_start (GTK_BOX (hbox), use_pw_chk, TRUE, TRUE, 0);

    set_passwd_btn = xfce_create_mixed_button_from_name (
        "linutop-lock", _("Set the password"));
    gtk_widget_set_sensitive (set_passwd_btn, FALSE);
    gtk_box_pack_start (GTK_BOX (hbox), set_passwd_btn, FALSE, FALSE, 0);

    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 2);

    g_signal_connect (G_OBJECT (use_pw_chk), "toggled",
        G_CALLBACK (on_pw_chk_toggled), (gpointer) set_passwd_btn);
    g_signal_connect (G_OBJECT (set_passwd_btn), "clicked",
        G_CALLBACK (on_pw_btn_clicked), NULL);

    view_only_chk = gtk_check_button_new_with_label (_(
        "Connected users can only view the desktop"));
    gtk_box_pack_start (GTK_BOX (vbox), view_only_chk, FALSE, FALSE, 2);

    allow_multi_chk = gtk_check_button_new_with_label (_(
        "Allow multiple connections"));
    gtk_box_pack_start (GTK_BOX (vbox), allow_multi_chk, FALSE, FALSE, 2);

    /* apply the config */
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (use_pw_chk), use_pw);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (view_only_chk), view_only);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (allow_multi_chk), allow_multi);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (autostart_chk), autostart);

    frame = xfce_framebox_new (_("VNC Configuration"), TRUE);
    gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_NONE);
    xfce_framebox_add (XFCE_FRAMEBOX (frame), vbox);
    gtk_box_pack_start (GTK_BOX (mhbox), frame, FALSE, FALSE, 0);

    /* separator */
    vs = gtk_separator_new (GTK_ORIENTATION_VERTICAL);
    gtk_box_pack_start (GTK_BOX (mhbox), vs, FALSE, FALSE, 2);

    /* server actions */
    vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 2);
    vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 2);

    bbox = gtk_button_box_new (GTK_ORIENTATION_VERTICAL);
    gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_START);
    gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 2);

    start_btn = xfce_create_mixed_button (GTK_STOCK_APPLY, _("Start"));
    g_object_set_data (G_OBJECT (start_btn), "id",
        GINT_TO_POINTER (SERVER_START));
    gtk_box_pack_start (GTK_BOX (bbox), start_btn, FALSE, FALSE, 2);
    restart_btn = xfce_create_mixed_button (GTK_STOCK_REFRESH, _("Restart"));
    g_object_set_data (G_OBJECT (restart_btn), "id",
        GINT_TO_POINTER (SERVER_RESTART));
    gtk_box_pack_start (GTK_BOX (bbox), restart_btn, FALSE, FALSE, 2);
    stop_btn = xfce_create_mixed_button (GTK_STOCK_STOP, _("Stop"));
    g_object_set_data (G_OBJECT (stop_btn), "id",
        GINT_TO_POINTER (SERVER_STOP));
    gtk_box_pack_start (GTK_BOX (bbox), stop_btn, FALSE, FALSE, 2);

    g_signal_connect (G_OBJECT (stop_btn), "clicked",
        G_CALLBACK (on_server_action_clicked), NULL);
    g_signal_connect (G_OBJECT (start_btn), "clicked",
        G_CALLBACK (on_server_action_clicked), NULL);
    g_signal_connect (G_OBJECT (restart_btn), "clicked",
        G_CALLBACK (on_server_action_clicked), NULL);

    set_action_buttons_visibility ();

    frame = xfce_framebox_new (_("Action"), TRUE);
    gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_NONE);
    xfce_framebox_add (XFCE_FRAMEBOX (frame), vbox);
    gtk_box_pack_start (GTK_BOX (mhbox), frame, FALSE, FALSE, 0);

    gtk_widget_show_all (window);

    g_object_set (gtk_settings_get_default (),
                  "gtk-button-images", TRUE, NULL);

    gtk_main ();

    g_list_free (pids);

    return 0;
}

